"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const path_1 = __importDefault(require("path"));
const cafs_1 = require("@pnpm/cafs");
const core_loggers_1 = require("@pnpm/core-loggers");
const error_1 = __importDefault(require("@pnpm/error"));
const graceful_fs_1 = __importDefault(require("@pnpm/graceful-fs"));
const logger_1 = __importDefault(require("@pnpm/logger"));
const package_is_installable_1 = __importDefault(require("@pnpm/package-is-installable"));
const read_package_json_1 = __importDefault(require("@pnpm/read-package-json"));
const dependency_path_1 = require("dependency-path");
const p_queue_1 = __importDefault(require("p-queue"));
const load_json_file_1 = __importDefault(require("load-json-file"));
const p_defer_1 = __importDefault(require("p-defer"));
const path_temp_1 = __importDefault(require("path-temp"));
const promise_share_1 = __importDefault(require("promise-share"));
const pick_1 = __importDefault(require("ramda/src/pick"));
const rename_overwrite_1 = __importDefault(require("rename-overwrite"));
const ssri_1 = __importDefault(require("ssri"));
const safeDeferredPromise_1 = __importDefault(require("./safeDeferredPromise"));
const TARBALL_INTEGRITY_FILENAME = 'tarball-integrity';
const packageRequestLogger = logger_1.default('package-requester');
const pickBundledManifest = pick_1.default([
    'bin',
    'bundledDependencies',
    'bundleDependencies',
    'dependencies',
    'directories',
    'engines',
    'name',
    'optionalDependencies',
    'os',
    'peerDependencies',
    'peerDependenciesMeta',
    'scripts',
    'version',
]);
function default_1(opts) {
    var _a;
    opts = opts || {};
    const networkConcurrency = (_a = opts.networkConcurrency) !== null && _a !== void 0 ? _a : 16;
    const requestsQueue = new p_queue_1.default({
        concurrency: networkConcurrency,
    });
    requestsQueue['counter'] = 0; // eslint-disable-line
    requestsQueue['concurrency'] = networkConcurrency; // eslint-disable-line
    const cafsDir = path_1.default.join(opts.storeDir, 'files');
    const getFilePathInCafs = cafs_1.getFilePathInCafs.bind(null, cafsDir);
    const fetch = fetcher.bind(null, opts.fetchers, opts.cafs);
    const fetchPackageToStore = fetchToStore.bind(null, {
        checkFilesIntegrity: cafs_1.checkFilesIntegrity.bind(null, cafsDir),
        fetch,
        fetchingLocker: new Map(),
        getFilePathByModeInCafs: cafs_1.getFilePathByModeInCafs.bind(null, cafsDir),
        getFilePathInCafs,
        requestsQueue,
        storeDir: opts.storeDir,
        verifyStoreIntegrity: opts.verifyStoreIntegrity,
    });
    const requestPackage = resolveAndFetch.bind(null, {
        engineStrict: opts.engineStrict,
        nodeVersion: opts.nodeVersion,
        pnpmVersion: opts.pnpmVersion,
        force: opts.force,
        fetchPackageToStore,
        requestsQueue,
        resolve: opts.resolve,
        storeDir: opts.storeDir,
        verifyStoreIntegrity: opts.verifyStoreIntegrity,
    });
    return Object.assign(requestPackage, { fetchPackageToStore, requestPackage });
}
exports.default = default_1;
async function resolveAndFetch(ctx, wantedDependency, options) {
    var _a, _b, _c, _d, _e, _f;
    let latest;
    let manifest;
    let normalizedPref;
    let resolution = (_a = options.currentPkg) === null || _a === void 0 ? void 0 : _a.resolution;
    let pkgId = (_b = options.currentPkg) === null || _b === void 0 ? void 0 : _b.id;
    const skipResolution = resolution && !options.update;
    let forceFetch = false;
    let updated = false;
    let resolvedVia;
    // When fetching is skipped, resolution cannot be skipped.
    // We need the package's manifest when doing `lockfile-only` installs.
    // When we don't fetch, the only way to get the package's manifest is via resolving it.
    //
    // The resolution step is never skipped for local dependencies.
    if (!skipResolution || options.skipFetch === true || Boolean(pkgId === null || pkgId === void 0 ? void 0 : pkgId.startsWith('file:')) || wantedDependency.optional === true) {
        const resolveResult = await ctx.requestsQueue.add(async () => ctx.resolve(wantedDependency, {
            alwaysTryWorkspacePackages: options.alwaysTryWorkspacePackages,
            defaultTag: options.defaultTag,
            lockfileDir: options.lockfileDir,
            preferredVersions: options.preferredVersions,
            preferWorkspacePackages: options.preferWorkspacePackages,
            projectDir: options.projectDir,
            registry: options.registry,
            workspacePackages: options.workspacePackages,
        }), { priority: options.downloadPriority });
        manifest = resolveResult.manifest;
        latest = resolveResult.latest;
        resolvedVia = resolveResult.resolvedVia;
        // If the integrity of a local tarball dependency has changed,
        // the local tarball should be unpacked, so a fetch to the store should be forced
        forceFetch = Boolean((((_c = options.currentPkg) === null || _c === void 0 ? void 0 : _c.resolution) != null) &&
            (pkgId === null || pkgId === void 0 ? void 0 : pkgId.startsWith('file:')) &&
            ((_d = options.currentPkg) === null || _d === void 0 ? void 0 : _d.resolution['integrity']) !== resolveResult.resolution['integrity'] // eslint-disable-line @typescript-eslint/dot-notation
        );
        updated = pkgId !== resolveResult.id || !resolution || forceFetch;
        resolution = resolveResult.resolution;
        pkgId = resolveResult.id;
        normalizedPref = resolveResult.normalizedPref;
    }
    const id = pkgId;
    if (resolution.type === 'directory') {
        if (manifest == null) {
            throw new Error(`Couldn't read package.json of local dependency ${wantedDependency.alias ? wantedDependency.alias + '@' : ''}${(_e = wantedDependency.pref) !== null && _e !== void 0 ? _e : ''}`);
        }
        return {
            body: {
                id,
                isLocal: true,
                manifest,
                normalizedPref,
                resolution: resolution,
                resolvedVia,
                updated,
            },
        };
    }
    const isInstallable = (ctx.force === true ||
        (manifest == null
            ? undefined
            : package_is_installable_1.default(id, manifest, {
                engineStrict: ctx.engineStrict,
                lockfileDir: options.lockfileDir,
                nodeVersion: ctx.nodeVersion,
                optional: wantedDependency.optional === true,
                pnpmVersion: ctx.pnpmVersion,
            })));
    // We can skip fetching the package only if the manifest
    // is present after resolution
    if ((options.skipFetch === true || isInstallable === false) && (manifest != null)) {
        return {
            body: {
                id,
                isLocal: false,
                isInstallable: isInstallable !== null && isInstallable !== void 0 ? isInstallable : undefined,
                latest,
                manifest,
                normalizedPref,
                resolution,
                resolvedVia,
                updated,
            },
        };
    }
    const fetchResult = ctx.fetchPackageToStore({
        fetchRawManifest: true,
        force: forceFetch,
        lockfileDir: options.lockfileDir,
        pkg: {
            ...pick_1.default(['name', 'version'], (_f = manifest !== null && manifest !== void 0 ? manifest : options.currentPkg) !== null && _f !== void 0 ? _f : {}),
            id,
            resolution,
        },
    });
    return {
        body: {
            id,
            isLocal: false,
            isInstallable: isInstallable !== null && isInstallable !== void 0 ? isInstallable : undefined,
            latest,
            manifest,
            normalizedPref,
            resolution,
            resolvedVia,
            updated,
        },
        bundledManifest: fetchResult.bundledManifest,
        files: fetchResult.files,
        filesIndexFile: fetchResult.filesIndexFile,
        finishing: fetchResult.finishing,
    };
}
function fetchToStore(ctx, opts) {
    const targetRelative = dependency_path_1.depPathToFilename(opts.pkg.id, opts.lockfileDir);
    const target = path_1.default.join(ctx.storeDir, targetRelative);
    if (!ctx.fetchingLocker.has(opts.pkg.id)) {
        const bundledManifest = p_defer_1.default();
        const files = p_defer_1.default();
        const finishing = p_defer_1.default();
        const filesIndexFile = opts.pkg.resolution['integrity']
            ? ctx.getFilePathInCafs(opts.pkg.resolution['integrity'], 'index')
            : path_1.default.join(target, 'integrity.json');
        doFetchToStore(filesIndexFile, bundledManifest, files, finishing); // eslint-disable-line
        if (opts.fetchRawManifest) {
            ctx.fetchingLocker.set(opts.pkg.id, {
                bundledManifest: removeKeyOnFail(bundledManifest.promise),
                files: removeKeyOnFail(files.promise),
                filesIndexFile,
                finishing: removeKeyOnFail(finishing.promise),
            });
        }
        else {
            ctx.fetchingLocker.set(opts.pkg.id, {
                files: removeKeyOnFail(files.promise),
                filesIndexFile,
                finishing: removeKeyOnFail(finishing.promise),
            });
        }
        // When files resolves, the cached result has to set fromStore to true, without
        // affecting previous invocations: so we need to replace the cache.
        //
        // Changing the value of fromStore is needed for correct reporting of `pnpm server`.
        // Otherwise, if a package was not in store when the server started, it will be always
        // reported as "downloaded" instead of "reused".
        files.promise.then((cache) => {
            // If it's already in the store, we don't need to update the cache
            if (cache.fromStore) {
                return;
            }
            const tmp = ctx.fetchingLocker.get(opts.pkg.id);
            // If fetching failed then it was removed from the cache.
            // It is OK. In that case there is no need to update it.
            if (tmp == null)
                return;
            ctx.fetchingLocker.set(opts.pkg.id, {
                ...tmp,
                files: Promise.resolve({
                    ...cache,
                    fromStore: true,
                }),
            });
        })
            .catch(() => {
            ctx.fetchingLocker.delete(opts.pkg.id);
        });
    }
    const result = ctx.fetchingLocker.get(opts.pkg.id);
    if (opts.fetchRawManifest && (result.bundledManifest == null)) {
        result.bundledManifest = removeKeyOnFail(result.files.then(async ({ filesIndex }) => {
            const { integrity, mode } = filesIndex['package.json'];
            const manifestPath = ctx.getFilePathByModeInCafs(integrity, mode);
            return readBundledManifest(manifestPath);
        }));
    }
    return {
        bundledManifest: (result.bundledManifest != null) ? promise_share_1.default(result.bundledManifest) : undefined,
        files: promise_share_1.default(result.files),
        filesIndexFile: result.filesIndexFile,
        finishing: promise_share_1.default(result.finishing),
    };
    async function removeKeyOnFail(p) {
        try {
            return await p;
        }
        catch (err) {
            ctx.fetchingLocker.delete(opts.pkg.id);
            throw err;
        }
    }
    async function doFetchToStore(filesIndexFile, bundledManifest, files, finishing) {
        try {
            const isLocalTarballDep = opts.pkg.id.startsWith('file:');
            if (!opts.force &&
                (!isLocalTarballDep ||
                    await tarballIsUpToDate(opts.pkg.resolution, target, opts.lockfileDir) // eslint-disable-line
                )) {
                let pkgFilesIndex;
                try {
                    pkgFilesIndex = await load_json_file_1.default(filesIndexFile);
                }
                catch (err) {
                    // ignoring. It is fine if the integrity file is not present. Just refetch the package
                }
                // if target exists and it wasn't modified, then no need to refetch it
                if ((pkgFilesIndex === null || pkgFilesIndex === void 0 ? void 0 : pkgFilesIndex.files) != null) {
                    const manifest = opts.fetchRawManifest
                        ? safeDeferredPromise_1.default()
                        : undefined;
                    if (pkgFilesIndex.name != null && opts.pkg.name != null && pkgFilesIndex.name !== opts.pkg.name ||
                        pkgFilesIndex.version != null && opts.pkg.version != null && pkgFilesIndex.version !== opts.pkg.version) {
                        /* eslint-disable @typescript-eslint/restrict-template-expressions */
                        throw new error_1.default('UNEXPECTED_PKG_CONTENT_IN_STORE', `\
Package name mismatch found while reading ${JSON.stringify(opts.pkg.resolution)} from the store. \
This means that the lockfile is broken. Expected package: ${opts.pkg.name}@${opts.pkg.version}. \
Actual package in the store by the given integrity: ${pkgFilesIndex.name}@${pkgFilesIndex.version}.`);
                        /* eslint-enable @typescript-eslint/restrict-template-expressions */
                    }
                    const verified = await ctx.checkFilesIntegrity(pkgFilesIndex.files, manifest);
                    if (verified) {
                        files.resolve({
                            filesIndex: pkgFilesIndex.files,
                            fromStore: true,
                            sideEffects: pkgFilesIndex.sideEffects,
                        });
                        if (manifest != null) {
                            manifest()
                                .then((manifest) => bundledManifest.resolve(pickBundledManifest(manifest)))
                                .catch(bundledManifest.reject);
                        }
                        finishing.resolve(undefined);
                        return;
                    }
                    packageRequestLogger.warn({
                        message: `Refetching ${target} to store. It was either modified or had no integrity checksums`,
                        prefix: opts.lockfileDir,
                    });
                }
            }
            // We fetch into targetStage directory first and then fs.rename() it to the
            // target directory.
            // Tarballs are requested first because they are bigger than metadata files.
            // However, when one line is left available, allow it to be picked up by a metadata request.
            // This is done in order to avoid situations when tarballs are downloaded in chunks
            // As much tarballs should be downloaded simultaneously as possible.
            const priority = (++ctx.requestsQueue['counter'] % ctx.requestsQueue['concurrency'] === 0 ? -1 : 1) * 1000; // eslint-disable-line
            const fetchManifest = opts.fetchRawManifest
                ? safeDeferredPromise_1.default()
                : undefined;
            if (fetchManifest != null) {
                fetchManifest()
                    .then((manifest) => bundledManifest.resolve(pickBundledManifest(manifest)))
                    .catch(bundledManifest.reject);
            }
            const fetchedPackage = await ctx.requestsQueue.add(async () => ctx.fetch(opts.pkg.id, opts.pkg.resolution, {
                lockfileDir: opts.lockfileDir,
                manifest: fetchManifest,
                onProgress: (downloaded) => {
                    core_loggers_1.fetchingProgressLogger.debug({
                        downloaded,
                        packageId: opts.pkg.id,
                        status: 'in_progress',
                    });
                },
                onStart: (size, attempt) => {
                    core_loggers_1.fetchingProgressLogger.debug({
                        attempt,
                        packageId: opts.pkg.id,
                        size,
                        status: 'started',
                    });
                },
            }), { priority });
            const filesIndex = fetchedPackage.filesIndex;
            // Ideally, files wouldn't care about when integrity is calculated.
            // However, we can only rename the temp folder once we know the package name.
            // And we cannot rename the temp folder till we're calculating integrities.
            const integrity = {};
            await Promise.all(Object.keys(filesIndex)
                .map(async (filename) => {
                const { checkedAt, integrity: fileIntegrity, } = await filesIndex[filename].writeResult;
                integrity[filename] = {
                    checkedAt,
                    integrity: fileIntegrity.toString(),
                    mode: filesIndex[filename].mode,
                    size: filesIndex[filename].size,
                };
            }));
            await writeJsonFile(filesIndexFile, {
                name: opts.pkg.name,
                version: opts.pkg.version,
                files: integrity,
            });
            if (isLocalTarballDep && opts.pkg.resolution['integrity']) { // eslint-disable-line @typescript-eslint/dot-notation
                await fs_1.promises.mkdir(target, { recursive: true });
                await graceful_fs_1.default.writeFile(path_1.default.join(target, TARBALL_INTEGRITY_FILENAME), opts.pkg.resolution['integrity'], 'utf8'); // eslint-disable-line @typescript-eslint/dot-notation
            }
            files.resolve({
                filesIndex: integrity,
                fromStore: false,
            });
            finishing.resolve(undefined);
        }
        catch (err) {
            files.reject(err);
            if (opts.fetchRawManifest) {
                bundledManifest.reject(err);
            }
        }
    }
}
async function writeJsonFile(filePath, data) {
    const targetDir = path_1.default.dirname(filePath);
    // TODO: use the API of @pnpm/cafs to write this file
    // There is actually no need to create the directory in 99% of cases.
    // So by using cafs API, we'll improve performance.
    await fs_1.promises.mkdir(targetDir, { recursive: true });
    const temp = path_temp_1.default(targetDir);
    await graceful_fs_1.default.writeFile(temp, JSON.stringify(data));
    await rename_overwrite_1.default(temp, filePath);
}
async function readBundledManifest(pkgJsonPath) {
    return pickBundledManifest(await read_package_json_1.default(pkgJsonPath));
}
async function tarballIsUpToDate(resolution, pkgInStoreLocation, lockfileDir) {
    let currentIntegrity;
    try {
        currentIntegrity = (await graceful_fs_1.default.readFile(path_1.default.join(pkgInStoreLocation, TARBALL_INTEGRITY_FILENAME), 'utf8'));
    }
    catch (err) {
        return false;
    }
    if (resolution.integrity && currentIntegrity !== resolution.integrity)
        return false;
    const tarball = path_1.default.join(lockfileDir, resolution.tarball.slice(5));
    const tarballStream = fs_1.createReadStream(tarball);
    try {
        return Boolean(await ssri_1.default.checkStream(tarballStream, currentIntegrity));
    }
    catch (err) {
        return false;
    }
}
async function fetcher(fetcherByHostingType, cafs, packageId, resolution, opts) {
    var _a, _b;
    const fetch = fetcherByHostingType[(_a = resolution.type) !== null && _a !== void 0 ? _a : 'tarball'];
    if (!fetch) {
        throw new Error(`Fetching for dependency type "${(_b = resolution.type) !== null && _b !== void 0 ? _b : 'undefined'}" is not supported`);
    }
    try {
        return await fetch(cafs, resolution, opts);
    }
    catch (err) {
        packageRequestLogger.warn({
            message: `Fetching ${packageId} failed!`,
            prefix: opts.lockfileDir,
        });
        throw err;
    }
}
//# sourceMappingURL=packageRequester.js.map